// Arduino版 Pongゲーム(Fck=16MHz)
// DELAY1/DELAY2は「プレイヤー側のラケットにボールが当たってからコンピュータ側のラケットが移動を開始するまでの遅延」
// DELAY1=15...(プレイヤー>コンピュータ)の時はコンピュータが強く
// DELAY2=16...(プレイヤー<=コンピュータ)の時はコンピュータが弱く
// 2006.12.19-2016.2.24 koyama@hirosaki-u.ac.jp
#undef  PROGMEM
#define PROGMEM __attribute__((section(".progmem.vars")))
#include <avr/sleep.h>
#define	DELAY1	15
#define	DELAY2	16
#define	SCREEN_W	48	// 48, 64, 96
#define	SCREEN_H	32
#define	LINEtop	72
#define	LINEdup	4
#define	LINEext	(LINEtop+SCREEN_H*LINEdup)
#define	SCROLL	2
#define	FONT_W	8
#define	FONT_H	16
#define	SYN		0x00
#define	BLK		0x10
#define	WHT		0x30

#if defined(ARDUINO_AVR_ATMEGA8O83)
// Nanoboard
#undef  TIMSK1
#define TIMSK1  TIMSK
#define	VR		6
#define Fck		8
#define BP		12
#define VS		90
#define HS		7
#define pxl(x)  {if(x){PORTB=WHT;}else{PORTB=BLK;nop();nop();}}

#elif defined(ARDUINO_AVR_PRO)
// Studuino
#define	VR		4
#define	Fck		8
#define	BP		15
#define	VS		90
#define	HS		7
#define pxl(x)  {if(x){PORTB=WHT;}else{PORTB=BLK;nop();nop();}}

#else
// UNO etc
#define	VR		0
#define	Fck		16
#define	BP		30
#define	VS		190
#define	HS		18
#define	pxl(x)	{if(x){PORTB=WHT;nop();nop();nop();nop();nop();nop();nop();}else{PORTB=BLK;nop();nop();nop();nop();nop();nop();nop();nop();nop();}}
#endif

#define	RSIZE	6
#define	R1x		43
#define	R2x		4
#define	max(x,y)	((x>y)?(x):(y))
#define	min(x,y)	((x<y)?(x):(y))
#define	nop()	asm volatile("nop"::)
#define	pxl8(x)	{pxl(x&0x80); pxl(x&0x40); pxl(x&0x20); pxl(x&0x10); pxl(x&0x08); pxl(x&0x04); pxl(x&0x02); pxl(x&0x01);}
uint8_t	vram[SCREEN_W/8*SCREEN_H];	// 48×32ドットのvram
#include "vlibnums.h"
volatile int ToneVsyncs;

ISR(TIMER1_COMPA_vect){}			// 64us

//------------------------------------------------------------------
// Simple tone generation
//	freq:	the frequency of the tone
//	dur:	The duration to play the tone in ms
#if !defined(ARDUINO_AVR_ATMEGA8O83)
void TVtone(unsigned int freq, unsigned int dur){
	if(freq==0)	return;
	TCCR2A = _BV(WGM21);	TCCR2B = _BV(CS20);
	uint8_t pre = 0b001;
	uint32_t ocr = 0;
	DDRB |= 0x08; //set pb3 (digital pin 11) to output
	//we are using an 8 bit timer, scan through prescalars to find the best fit
	ocr=F_CPU/freq/2-1;	pre=0b001;			// ck/1: same for both timers
	if(ocr>255){
		ocr=F_CPU/freq/2/8-1;	pre=0b010;	// ck/8: same for both timers
		if(ocr>255){	ocr=F_CPU/freq/2/32-1;	pre=0b011;}
		if(ocr>255){
			ocr=F_CPU/freq/2/64-1;	pre=0b100;
			if(ocr>255){ocr=F_CPU/freq/2/128-1;	pre=0b101;}
			if(ocr>255){
				ocr=F_CPU/freq/2/256-1;	pre=0b110;
				if(ocr>255){ocr=F_CPU/freq/2/1024-1;	pre=0b111;}
			}
		}
	}
	TCCR2B = pre;
	if(dur>0)	ToneVsyncs = dur*60/1000;	// 60: framerate
	else		ToneVsyncs = -1;
	OCR2A = ocr;				// Set the OCR for the given timer,
	//set it to toggle the pin by itself
	TCCR2A &= ~(_BV(COM2A1));	//set COM2A1 to 0
	TCCR2A |= _BV(COM2A0);
}

/* Stops tone generation */
void TVnoTone() {
	TCCR2B = 0;
	PORTB &= 0xf7;				//set pin 11 to 0
}
#endif
//---------------------------------------------------------
uint8_t analogRead8(uint8_t pin){
  ADMUX = (0x01<<6) | (1<<ADLAR) | (pin&0x07);
  ADCSRA= (1<<ADEN) | (1<<ADSC)  | 0x02;
  while(bit_is_set(ADCSRA, ADSC));
  return ADCH;
}
//---------------------------------------------------------
void setup(){
	uint8_t	b0,b1,b2,b3,b4,i;
	uint8_t	*p;						// vram[]のポインタ
	uint8_t	line=0;					// 走査線カウンタで64us毎にカウントアップ
	uint8_t	fcntr=0;				// フレームカウンタ
	uint8_t	S=0;					// スイッチ操作に伴う状態
	uint8_t	x=R2x, y;				// ボールの位置
	int8_t	r1y=(SCREEN_H-RSIZE)/2, r2y=(SCREEN_H-RSIZE)/2;	// ラケットのy座標
	int8_t	r2y_;					// ラケット2の予想位置
	uint8_t score1=0, score2=0;		// スコアとスコア表示
	int8_t	dx=1, dy=1;				// ボールの移動差分
	PORTB=0x00;		DDRB=0x38;		// oooxxx
	ICR1=64*Fck/8-1; TCCR1A=0x00;  TCCR1B=0x1a;  TCNT1=0;// CTC on ICR1=127(64us),CK/8
	Vline(0,0,SCREEN_W-1,0,1);	Vline(SCREEN_W-1,0,SCREEN_W-1,SCREEN_H-1,1);
	Vline(SCREEN_W-1,SCREEN_H-1,0,SCREEN_H-1,1);	Vline(0,SCREEN_H-1,0,0,1);
	Vline(R1x,r1y, R1x,r1y+RSIZE-1,2);		// ラケット1表示
	Vline(R2x,r2y, R2x,r2y+RSIZE-1,2);		// ラケット2表示
#if defined(ARDUINO_AVR_ATMEGA8O83)
	TIMSK=(1<<OCIE1A);
#else
	TIMSK0=0x00;
	TIMSK1|=(1<<OCIE1A);
#endif
	for(;;){
		set_sleep_mode(SLEEP_MODE_IDLE);sleep_mode();	// SE=1, sleep, SE=0
		// タイマー割り込みでスリープからwake up
		PORTB=SYN;	for(i=0;i<HS;i++)	nop();			// H-Sync
		if(++line<9){
			for(i=0;i<VS;i++)	nop();	PORTB=BLK;		// V-Sync
			p=vram;
		}else{
			PORTB=BLK;
			if(line>=LINEtop && line<LINEext){			// ビデオ信号出力
				for(i=0;i<BP;i++)	PORTB=BLK;
				b0=*p;b1=*(p+1);b2=*(p+2);b3=*(p+3);b4=*(p+4);i=*(p+5);
				pxl8(b0);pxl8(b1);pxl8(b2);pxl8(b3);pxl8(b4);pxl8(i);
				nop();nop();PORTB=BLK;
				if(line%LINEdup==LINEdup-1)	p+=(SCREEN_W/8);// ライン重複
			}else if(line==LINEext&&S==0&&r1y==0){			// ラケット上端でスタート
				score1=score2=0;							// スコア初期化
				fcntr=0;
				S=1;
			}else if(line==LINEext+1&&++fcntr==61&&S==1){	//1秒後にプレイ開始
				dx=dy=1;	if((r1y+r2y)&1)	dy=-1;
				x=R2x+1;	y=r2y+RSIZE/2;
				Vplot(x,y,1);								// ボール表示
				S=2;
 			}else if(line==LINEext+3 && fcntr%2==0 && S==2){// ボールの表示更新
				Vplot(x,y,2);								// 現在位置のボール消去
				x+=dx;
				if(x==0)				score1++;			// Point of Player1
				else if(x==SCREEN_W-1)	score2++;			// Point of Player2
				else if(x==R1x&&y>=r1y&&y<=r1y+RSIZE){		// ラケット1にHit
					dx=-dx;
#if defined(ARDUINO_AVR_ATMEGA8O83)
					ToneVsyncs = 2*96;						// 200ms
#else
					TVtone(440, 100);
#endif
				}else if(x==R2x&&y>=r2y&&y<=r2y+RSIZE){		// ラケット2にHit
					dx=-dx;
#if defined(ARDUINO_AVR_ATMEGA8O83)
					ToneVsyncs = 2*96;						// 200ms
#else
					TVtone(440, 100);
#endif
				}
				y+=dy;
				if(y==0 || y==SCREEN_H-1 || x==R1x&&(dy==1&&y==r1y||dy==-1&&y==r1y+RSIZE-1))	dy=-dy; // y方向転換
				if(x==0 || x==SCREEN_W-1){
					if(score1==9 || score2==9)	S=0;		// ゲームオーバー
					else{	fcntr=0;	S=1;	}			// 1秒後にプレイ開始
				}else	Vplot(x,y,2);						// ボール表示
			}else if(line==LINEext+4){						// ラケット1の更新
				Vline(R1x,r1y, R1x,r1y+RSIZE-1,2);			// ラケット1消去
				r1y=min(SCREEN_H-RSIZE-1,max(0,SCREEN_H-1-(analogRead8(VR)>>3)));
			}else if(line==LINEext+5){
				Vline(R1x,r1y, R1x,r1y+RSIZE-1,2);			// ラケット1表示
			}else if(line==LINEext+6){
				Vline(R2x,r2y, R2x,r2y+RSIZE-1,2);			// ラケット2消去
			}else if(line==LINEext+7 && fcntr%4==0 && S==2 && dx==-1 && x<R1x-((score1>score2)?DELAY1:DELAY2)){
				if(dy==1){
					r2y_=y+x-R2x;
					if(r2y_>=SCREEN_H)	r2y_=SCREEN_H-(r2y_-SCREEN_H);
				}else{
					r2y_=y-x+R2x;
					if(r2y_<0)	r2y_=-r2y_;
				}
			}else if(line==LINEext+8 && fcntr%4==0 && S==2 && dx==-1 && x<R1x-((score1>score2)?DELAY1:DELAY2)){
				if(r2y+RSIZE/2>r2y_ && r2y>0)					r2y--;
				if(r2y+RSIZE/2<r2y_ && r2y<SCREEN_H-RSIZE-1)	r2y++;
			}else if(line==LINEext+9){
				Vline(R2x,r2y, R2x,r2y+RSIZE-1,2);			// ラケット2表示
			}else if(line==LINEext+10){
				Vputc4x6(SCREEN_W/2+4,2,'0'+score1);		// スコア1表示
				Vputc4x6(SCREEN_W/2-8,2,'0'+score2);		// スコア2表示
#if defined(ARDUINO_AVR_ATMEGA8O83)
			}
			if(line%16==0 && ToneVsyncs>0){	ToneVsyncs--;	PORTB^=0x08;	}
#else
			}else if(line==LINEext+11){
				if(ToneVsyncs != 0){
					if(ToneVsyncs>0)	ToneVsyncs--;
				}else{
					TCCR2B=0;	// stop the tone
					PORTB &= 0xf7;
				}
			}
#endif
		}
	}
}
void loop(){}
